import matplotlib
matplotlib.rc('contour', negative_linestyle='solid')

from matplotlib.backends.backend_agg import \
    FigureCanvasAgg

from matplotlib.backends.backend_pdf import \
    FigureCanvasPdf

from matplotlib.backends.backend_ps import \
    FigureCanvasPS

from matplotlib.backends.backend_svg import \
    FigureCanvasSVG
